#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Basic Libraries
import asyncio as aio
import sys
import time
from threading import Thread
from typing import Optional, List, Callable
# NDN Imports
from ndn.app import NDNApp
from ndn.encoding import Name, BinaryStr
from ndn.security import Keychain
from ndn.transport.stream_socket import Face
from ndn.storage import Storage, DiskStorage
# Custom Imports
from .core import SVSyncCore
from .logger import SVSyncLogger
from .missing_data import MissingData
from .security import SecurityOptions
from .svs_base import SVSyncBase

# Class Type: an abstract API thread class
# Class Purpose:
#   to push SVS to a separate thread so SVS does not become encounter a block.
#   to derive different SVSync_Thread classes from.
#   to allow the user to interact with SVS, fetch and publish.
class SVSyncBase_Thread(Thread):
    class SVSyncUnwaitedThread(Exception):
        pass
    def __init__(self, groupPrefix:Name, nid:Name, updateCallback:Callable, storage:Optional[Storage]=None, securityOptions:Optional[SecurityOptions]=None, face:Optional[Face]=None, keychain:Optional[Keychain]=None) -> None:
        SVSyncLogger.info("SVSync_Thread: Created thread to push SVS to.")
        Thread.__init__(self)
        self.groupPrefix, self.nid, self.updateCallback, self.storage, self.face, self.keychain, self.secOptions, self.svs, self.loop, self.app, self.failed = groupPrefix, nid, updateCallback, storage, face, keychain, securityOptions, None, None, None, False
    def wait(self) -> None:
        while self.svs is None:
            time.sleep(0.001)
            if self.failed:
                sys.exit()
    def run(self) -> None:
        def loop_task() -> None:
            self.app = NDNApp(self.face, self.keychain)
            if issubclass(type(self.storage), DiskStorage):
                self.storage.initialize()
            try:
                self.app.run_forever(after_start=self.function())
            except (FileNotFoundError, ConnectionRefusedError):
                print("Error: could not connect to NFD for SVS.")
                self.failed = True
                sys.exit()

        self.loop = aio.new_event_loop()
        aio.set_event_loop(self.loop)
        self.loop.create_task(loop_task())
        self.loop.run_forever()
    async def function(self) -> None:
        raise NotImplementedError
    def missing_callback(self, missing_list:List[MissingData]) -> None:
        aio.get_event_loop().create_task(self.updateCallback(self)(missing_list))
    def getSVSync(self) -> Optional[SVSyncBase]:
        return self.svs
    async def fetchData(self, nid:Name, seqno:int, retries:int=0) -> Optional[bytes]:
        try:
            data = await self.svs.fetchData(nid, seqno, retries)
            return data
        except AttributeError:
            raise self.SVSyncUnwaitedThread("A SVSync Thread needs to be waited on before doing operations.")
    async def fetchDataPacket(self, nid:Name, seqno:int, retries:int=0) -> Optional[BinaryStr]:
        try:
            pck = await self.svs.fetchDataPacket(nid, seqno, retries)
            return pck
        except AttributeError:
            raise self.SVSyncUnwaitedThread("A SVSync Thread needs to be waited on before doing operations.")
    def publishData(self, data:bytes) -> None:
        try:
            self.svs.publishData(data)
        except AttributeError:
            raise self.SVSyncUnwaitedThread("A SVSync Thread needs to be waited on before doing operations.")
    def getCore(self) -> Optional[SVSyncCore]:
        return self.svs.getCore() if self.svs else None


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Basic Libraries
from typing import Callable, Optional, Tuple
# NDN Imports
from ndn.app import NDNApp
from ndn.encoding import Name, BinaryStr, Component, parse_data
from ndn.types import InterestNack, InterestTimeout, InterestCanceled, ValidationFailure
from ndn.storage import Storage
# Custom Imports
from .logger import SVSyncLogger
from .security import SecurityOptions
from .svs_base import SVSyncBase

# Class Type: an derived API class
# Class Purpose:
#   to allow the user to interact with SVS, fetch and publish.
#   to allow caching other node's data in case one node goes down.
class SVSyncShared(SVSyncBase):
    def __init__(self, app:NDNApp, groupPrefix:Name, nid:Name, updateCallback:Callable, cacheOthers:bool, storage:Optional[Storage]=None, securityOptions:Optional[SecurityOptions]=None) -> None:
        self.cacheOthers = cacheOthers
        preDataPrefix = groupPrefix + [Component.from_str("data")] if self.cacheOthers else groupPrefix + [Component.from_str("data")] + nid
        preSyncPrefix = groupPrefix + [Component.from_str("sync")]
        super().__init__(app, preSyncPrefix, preDataPrefix, groupPrefix, nid, updateCallback, storage, securityOptions)
    async def _fetch(self, nid:Name, seqno:int, retries:int=0) -> Tuple[Optional[bytes], Optional[BinaryStr]]:
        name = self.getDataName(nid, seqno)
        while retries+1 > 0:
            try:
                SVSyncLogger.info(f'SVSync: fetching data {Name.to_str(name)}')
                _, _, _, pkt = await self.app.express_interest(name, need_raw_packet=True, must_be_fresh=True, can_be_prefix=False, lifetime=self.DATA_INTEREST_LIFETIME)
                ex_int_name, meta_info, content, sig_ptrs = parse_data(pkt)
                isValidated = await self.secOptions.validate(ex_int_name, sig_ptrs)
                if not isValidated:
                    return (None, None)
                SVSyncLogger.info(f'SVSync: received data {bytes(content)}')
                if content and self.cacheOthers:
                    self.storage.put_packet(name, pkt)
                return (bytes(content), pkt) if content else (None, pkt)
            except InterestNack as e:
                SVSyncLogger.warning(f'SVSync: nacked with reason={e.reason}')
            except InterestTimeout:
                SVSyncLogger.warning("SVSync: timeout")
            except InterestCanceled:
                SVSyncLogger.warning("SVSync: canceled")
            except ValidationFailure:
                SVSyncLogger.warning("SVSync: data failed to validate")
            except Exception as e:
                SVSyncLogger.warning(f'SVSync: unknown error has occured: {e}')

            retries = retries - 1
            if retries+1 > 0:
                SVSyncLogger.info("SVSync: retrying fetching data")
        return (None, None)
    async def fetchDataPacket(self, nid:Name, seqno:int, retries:int=0) -> Optional[BinaryStr]:
        _, pkt = await self._fetch(nid, seqno, retries)
        return pkt
    async def fetchData(self, nid:Name, seqno:int, retries:int=0) -> Optional[bytes]:
        data, _ = await self._fetch(nid, seqno, retries)
        return data
    def getDataName(self, nid:Name, seqno:int) -> Name:
        return ( self.groupPrefix + [Component.from_str("data")] + nid + [Component.from_str(str(seqno))] )
    def serveDataPacket(datapkt:BinaryStr) -> None:
        name, _, content, _ = parse_data(datapkt)
        if content:
            self.storage.put_packet(name, datapkt)


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Basic Libraries
from typing import Callable, Optional
# NDN Imports
from ndn.app import NDNApp
from ndn.encoding import Name, Component
from ndn.storage import Storage
# Custom Imports
from .security import SecurityOptions
from .svs_base import SVSyncBase

# Class Type: an API class
# Class Purpose:
#   to allow the user to interact with SVS, fetch and publish.
class SVSync(SVSyncBase):
    def __init__(self, app:NDNApp, groupPrefix:Name, nid:Name, updateCallback:Callable, storage:Optional[Storage]=None, securityOptions:Optional[SecurityOptions]=None) -> None:
        preSyncPrefix, preDataPrefix = groupPrefix + [Component.from_str("sync")], nid + groupPrefix + [Component.from_str("data")]
        super().__init__(app, preSyncPrefix, preDataPrefix, groupPrefix, nid, updateCallback, storage, securityOptions)
    def getDataName(self, nid:Name, seqno:int) -> Name:
        return ( nid + self.groupPrefix + [Component.from_str("data")] + [Component.from_str(str(seqno))] )


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# NDN Imports
from ndn.encoding import Component, TlvModel, UintField, BytesField, DecodeError
# Custom Imports
from .tlv import SVSyncTlvTypes

# Class Type: a tlv structure for encoding
# Class Purpose:
#   hold all aspects of MetaData in a Model
class MetaDataModel(TlvModel):
    source = BytesField(SVSyncTlvTypes.META_SOURCE.value)
    tseqno = UintField(SVSyncTlvTypes.META_TOTAL.value)
    nopcks = UintField(SVSyncTlvTypes.META_NOPCKS.value)

# Class Type: a API class
# Class Purpose:
#   to make metadata model eaiser to interact with
class MetaData:
    __slots__ = ('source','tseqno','nopcks')
    def __init__(self, comp:Component=None) -> None:
        try:
            model:MetaDataModel = MetaDataModel.parse(bytes(Component.get_value(comp)))
            self.source, self.tseqno, self.nopcks = model.source, model.tseqno, model.nopcks
        except (ValueError,TypeError,IndexError,DecodeError):
            self.source, self.tseqno, self.nopcks = b'', 0, 0
    def encode(self) -> Component:
        model:MetaDataModel = MetaDataModel()
        model.source, model.tseqno, model.nopcks = self.source, self.tseqno, self.nopcks
        return Component.from_bytes(model.encode())


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Basic Libraries
from enum import Enum

# Class Type: an enumeration struct
# Class Purpose:
#   to differ tlv model types.
class SVSyncTlvTypes(Enum):
    META_SOURCE   = 198
    META_TOTAL    = 199
    META_NOPCKS   = 200
    VECTOR        = 201
    VECTOR_ENTRY  = 202
    SEQNO         = 203